"""
To be implemented
"""
import numpy as np
from scipy.stats import gaussian_kde

def density_kernel(data, cov_factor=0.1, n_points=300):    
    density = gaussian_kde(data)
    xs = np.linspace(data.min(), data.max(), n_points)
    density.covariance_factor = lambda : cov_factor
    density._compute_covariance()
    return xs, density(xs)


"""
Stats plots (just a test)
"""
try:
    import statsmodels.api as sm
    import matplotlib.pyplot as plt
    import plotly.graph_objects as go
except:
    pass

from . import _preprocess
from ..plots.plotly_utils import get_figure

def qqplot(x, alpha=0.3, ms=20):
        pp = sm.ProbPlot(x, fit=True)
        qq = pp.qqplot(marker='.', markerfacecolor='k', markeredgecolor='k', alpha=alpha, markersize=ms)
        sm.qqline(qq.axes[0], line='45', fmt='k--')
        return plt.gcf()
    
def density_kernel(*X, cov_factor=0.1, n_points=300, **kwargs):
    fig = get_figure(yaxis_title="Probability density", **kwargs)
    for x in X:
        xs, density = _preprocess.density_kernel(x, cov_factor=cov_factor, n_points=n_points)
        
        fig.add_trace(go.Scatter(x=xs, y=density, mode="lines", showlegend=False, line_width=4))
    return fig


import numpy as np
from numpy.lib.stride_tricks import as_strided

def rolling_view(x, window):
    """
    Returns rolling view (no extra memory needed) over first axis of x.
    
    Output shape: ``(x.shape[0]-window+1, window, *x.shape[1:])``
    """
    stride = x.strides[0]
    shape = [x.shape[0] - window + 1, window]
    strides = [stride, stride]
    if x.ndim > 1:
        shape += [*x.shape[1:]]
        strides += [*x.strides[1:]]    
    return as_strided(x, shape=shape, strides=strides)

def idxs_condition(x, condition):
    """
    Returns idx where condition is first met. Ignore rest until fails and mets the condition again.
    
    Attributes:
        x:          np.ndarray
        condition:  f: x -> bool array
    """
    verify_condition = condition(x)
    changes = np.where(np.abs(np.diff((verify_condition).astype(int))) == 1)[0]
    if changes.size > 0:
        changes += 1
        if verify_condition[changes[0]]:
            return changes[::2]
        else:
            return changes[1::2]
    else:
        return changes
    
def numpy_fill(arr):
    '''Fills nan with previous not-NaN values'''
    mask = np.isnan(arr)
    if mask.all():
        return np.NaN
    else:
        idx = np.where(~mask, np.arange(mask.shape[0]), 0)
        np.maximum.accumulate(idx,axis=0, out=idx)
        out = arr[np.arange(idx.shape[0])[idx]]
        if mask[0] and idx[0] == 0:
            out[:(idx == 0).sum()] = arr[idx[idx > 0][0]]
        return out
    
def percentile_boot_paired(X, Y, func, alpha=0.05, R=int(1e4), seed=0, names=None):
    """"
    Bootstrap a statistic that takes X, Y as input. 
    NOTE: If the stat can be written in numpy, a modified numba version is recommended.
    """
    np.random.seed(seed)
    N = X.size    
    data_paired = np.vstack((X, Y)).T
    idxs_resampling = np.random.randint(low=0, high=N, size=R*N)
    data_resampled = data_paired[idxs_resampling].reshape(R, N, 2)
    
    stat = func(X, Y)
    if hasattr(stat, "__len__"):
        boot_sample = np.empty((R, len(stats)))
    else:
        boot_sample = np.empty((R))
    for i, r in enumerate(tqdm(data_resampled)):
        boot_sample[i] = func(*r.T)
    alpha_ptg = alpha*100
    results = pd.Series(dict(stat=names,
                             sample_stat=stat,
                             lower_bound=np.percentile(boot_sample, alpha_ptg, axis=0),
                             upper_bound=np.percentile(boot_sample, 100-alpha_ptg, axis=0),
                             CI=np.percentile(boot_sample, [alpha_ptg/2, 100 - alpha_ptg/2], axis=0),
                            ))
    return results


"""
Adaptation of resample R package: https://cran.r-project.org/web/packages/resample/resample.pdf
Based on the article: https://arxiv.org/abs/1411.5279
"""
import numpy as np
import pandas as pd
try:
    import rpy2.robjects as ro
    from rpy2.robjects import r, pandas2ri, numpy2ri
    from rpy2.robjects.packages import importr
    pandas2ri.activate()
    ro.numpy2ri.activate()
    from ._helper import attr_preprocess, load_R_pkg
except:
    pass

    
def stat_computer(obj_name, nsamples=1):
    load_R_pkg("data.table")
    if nsamples == 1:
        return r(f"{obj_name}$stats") 
    else:
        return dict(stats=r(f"{obj_name}$stats"),
                    individual_stats=r(f"rbindlist(lapply(lapply({obj_name}$resultsBoth, (function (l) l$stats)), as.data.frame.list), idcol=TRUE)").set_index(".id")
                   )

def bootstrap(x, method, y=None, stat="mean", N=int(1e5), seed=0, block_size=1000, **kwargs):
    """
    Attrs:
        - method: ["t", "percentile", "bca", "bootstrapT"]
        
    Returns: bootstrap object (R)
    """
    kwargs_str = attr_preprocess(bootstrap, locals())
    load_R_pkg("resample")
    if y is None: # one sample       
        if method == "bootstrapT":
            ro.globalenv["bs"] = r(f"bootstrap(x, c({stat} = {stat}(x), sd = sd(x)), R=N, seed=0, block.size=block.size {kwargs_str})")
        else:
            ro.globalenv["bs"] = r(f"bootstrap(x, c({stat} = {stat}(x)), R=N, seed=0, block.size=block.size {kwargs_str})")
    else:
        if method == "bootstrapT":
            raise ValueError(f"method {method} not implemented for the 2-sample case")
        else:
            ro.globalenv["DF"] = pd.concat([pd.DataFrame(dict(val=x)).assign(Group="x"),
                                            pd.DataFrame(dict(val=y)).assign(Group="y")
                                           ], ignore_index=True, axis=0)
            ro.globalenv["bs"] = r(f"bootstrap2(DF, {stat}(val), treatment=Group, R=N, seed=seed, block.size=block.size {kwargs_str})")
    return ro.globalenv["bs"]

def CI_bootstrap(x, y=None, stat="mean", method="bootstrapT", expand=True, alpha=0.05, probs=None, alternative="two-sided", return_stats=False, **kwargs):
    """
    Check for the stats. Unbiased, functional statistics such as the mean should have zero bootstrap bias. If not, increase N.
    NOTE: s2 will have non-zero bootstrap bias because it is not functional.
    
    Attrs:
        - method: ["t", "percentile", "bca", "bootstrapT"].
        - alternative: ["two-sided", "less", "greater"].
    """
    if probs is None:
        if alternative == "less":
            ro.globalenv["probs"] = np.array([1 - alpha])
        elif alternative == "greater":
            ro.globalenv["probs"] = np.array([alpha])
        elif alternative == "two-sided":
            ro.globalenv["probs"] = np.array([alpha/2, 1 - alpha/2])
        else:
            raise ValueError(f"alternative {alternative} not valid. Available: 'greater', 'less', 'two-sided'")
            
    bs = bootstrap(x, method, y=y, stat=stat, **kwargs)
    
    if method == "bootstrapT":
        CI = r(f"CI.{method}(bs, probs=probs)") # bootstrap T is not affected by narrowness bias. Table 6
    else:
        ro.globalenv["expand"] = expand
        CI = r(f"CI.{method}(bs, probs=probs, expand=expand)")
    if CI.shape[0] == 1:
        CI = CI[0]
    
    if probs is None:
        if alternative == "less":
            CI = np.array([-np.inf, CI[0]]) # Only implemented for one at a time.
        elif alternative == "greater":
            CI = np.array([CI[0], np.inf])
            
    if return_stats:
        return bs, stat_computer("bs", nsamples=1 + int(y is not None)), CI
    else:
        return CI
    
def permutation(x, y=None, stat="mean", N=int(1e5), seed=0, block_size=1000,  **kwargs):
    """
    Returns: permutationTest object (R) and the results of the test. 
    
    The implementation in numba is faster and includes pairing option, which this one does not.
    """
    kwargs_str = attr_preprocess(permutation, locals())
    load_R_pkg("resample")   
    if y is None: # one sample       
        ro.globalenv["perm"] = r(f"permutationTest(x, c({stat} = {stat}(x)), R=N, seed=0, block.size=block.size {kwargs_str})")
    else:
        ro.globalenv["DF"] = pd.concat([pd.DataFrame(dict(val=x)).assign(Group="x"),
                                        pd.DataFrame(dict(val=y)).assign(Group="y")
                                       ], ignore_index=True, axis=0)
        ro.globalenv["perm"] = r(f"permutationTest2(DF, {stat}(val), treatment=Group, R=N, seed=seed, block.size=block.size {kwargs_str})")
        
    return ro.globalenv["perm"], stat_computer("perm", nsamples= 1 + int(y is not None))



"""
Not implemented
"""
try:
    from statsmodels.stats.diagnostic import lilliefors
except:
    pass

def normality(df, path_dict={}, save=True):
    """Lilliefors test for each column"""
    S = pd.Series({col: lilliefors(df[col].values)[1] for col in df.columns})
    return S


from numba import njit
import numpy as np
import pandas as pd
import math
import warnings
from . import bootstrap
try:
    import statsmodels.stats.api as sms
except:
    pass

def t_interval(x, alpha=0.05, alternative="two-sided"):
    return sms.DescrStatsW(x).tconfint_mean(alpha=alpha, alternative=alternative)

@njit
def compute_coverage(CI, data, stat, N, seed=0, num_iters=1000, closed_right=True, closed_left=True, tol=1e-8):
    low, high = CI
    estimates = bootstrap.resample_nb(data, stat, R=num_iters, N=N, seed=seed)[:, 0]
    if closed_left:
        low_fails = (estimates < low).mean()
    else:
        low_fails = (estimates <= (low+tol)).mean()
    if closed_right:
        high_fails = (estimates > high).mean()
    else:
        high_fails = (estimates >= (high-tol)).mean()
    coverage = 1 - (low_fails + high_fails)
    return np.array([low_fails, high_fails, coverage])

def coverage(*args, num_N=20, **kwargs):
    Ns = np.unique(np.linspace(2, args[1].shape[0], num_N, dtype=int))
    covg_data = np.vstack([compute_coverage(*args, N=N, **kwargs) for N in Ns]).T # shape: (3, num_N).  3: low_fails, high_fails, coverage
    return Ns, covg_data

def CI_specs(CIs, data, stat, coverage_iters=int(1e4), seed=42, avg_len=3, **coverage_kws):
    CI_arr = CIs.values
    sample_stat = stat(data)
    coverages = np.stack([coverage(CI, data, stat, num_iters=coverage_iters, seed=seed, **coverage_kws)[1] for CI in CI_arr], axis=1)
    # coverages shape: (3, #CI, num_N)
    low_fails_last, high_fails_last, coverage_last = coverages[:, :, -1]
    low_fails_avg, high_fails_avg, coverage_avg = coverages[:, :, -avg_len:].mean(axis=-1)

    CIs2 = CIs.copy()
    CIs2['width'] = np.hstack([np.diff(CI) for CI in CI_arr])
    CIs2['asymmetry'] = np.hstack([(CI[1] - sample_stat) / (sample_stat - CI[0]) for CI in CI_arr])
    env = locals()    
    for end in ['avg', 'last']:
        for k in ['low_fails', 'high_fails', 'coverage']:
            key = f"{k}_{end}"
            CIs2[key.replace("_", "-")] = env[key]
    return CIs2

def find_best(CIs, data=None, stat=None, alpha=0.05, alternative=None, alpha_margin_last=0.0075, alpha_margin_avg=0.015, **kwargs):
    alpha_expanded_last = alpha + alpha_margin_last
    alpha_expanded_avg = alpha + alpha_margin_avg
    if 'coverage-last' not in CIs.columns:
        CIs = CI_specs(CIs, data, stat, **kwargs)
    if alternative is None:
        if (~np.isfinite(CIs['low'].values)).all():
            alternative = 'less'
        elif (~np.isfinite(CIs['high'].values)).all():
            alternative = 'greater'
        else:
            alternative = 'two-sided'
    
    coverages_last, coverages_avg = CIs[['coverage-last', 'coverage-avg']].values.T
    valid = np.unique(np.hstack([np.where(coverages_last >= (1-alpha_expanded_last))[0],
                                 np.where(coverages_avg >= (1-alpha_expanded_avg))[0],
                                 np.abs(coverages_last - (1-alpha)).argmin()]))
    if alternative == 'two-sided':
        best_interval = CIs['width'].values[valid].argmin()
    elif alternative == 'less':
        best_interval = CIs['high'].values[valid].argmin()
    elif alternative == 'greater':
        best_interval = CIs['low'].values[valid].argmax()
    return CIs.iloc[valid].iloc[best_interval]

def ci_percentile_equal_tailed(x, p, alpha=0.05, alternative='two-sided'):
    """
    Exact confidence interval for percentiles. 
    Attrs:
        - p:  percentile in (0, 1)
            
    Returns CI [Yi, Yj] such that Prob(x in CI) => 1 - alpha. Yi, Yj are order statistics.
    https://online.stat.psu.edu/stat415/book/export/html/835
    """
    n = x.size
    if n < 500:    
        def binom_cdf(x): 
            return np.array([math.comb(n, k) * p**k * (1-p)**(n-k) for k in range(x+1)]).sum()
        binom_cdf = np.vectorize(binom_cdf)
        #def binom_pmf(x):    
        #    return math.comb(n, x) * p**x * (1-p)**(n-x)
    else:
        try:
            from scipy.stats import binom
            binom_cdf = lambda x: binom.cdf(x, n, p)
        except:
            raise ImportError('scipy not found. Please install it for n > 500. Alternatively, use the normal aproximation for the binomial.')          
    
    p_below_percentile = binom_cdf(np.arange(n+1))
    
    if alternative == 'two-sided':
        lows = np.where(p_below_percentile <= alpha/2)[0]
        if lows.size > 0:
            l = lows[-1] + 1 
        else:
            warnings.warn('n is too small to warrantee an exact lower bound other than the minimum.', RuntimeWarning)
            l = 0
        uppers = np.where(p_below_percentile >= (1- alpha/2))[0]
        if uppers.size > 0:
            u = uppers[0] - 1
        else:
            warnings.warn('n is too small to warrantee an exact upper end other than the maximum.', RuntimeWarning)
            u = -1
        ci_equal_tailed = [l, u]
        if l == 0:
            q0 = 0
            CI_prob0 = 0
        else:
            q0 = p_below_percentile[l-1] #  p(X < l) = p(x <= l-1) = cdf(l-1)
            CI_prob0 = l/n # add 1 to both endpoints (indexing in python starts at 0)
        if u == -1:
            q1 = 1
            CI_prob1 = 1
        else:
            q1 = p_below_percentile[u+1]
            CI_prob1 = (u+1) / n # add 1 to both endpoints (indexing in python starts at 0)
        quantiles = np.array([q0, q1]) 
        CI_prob = np.array([CI_prob0, CI_prob1])
        CI = np.sort(x)[ci_equal_tailed]
    elif alternative == 'less':
        uppers = np.where(p_below_percentile >= (1-alpha))[0]
        if uppers.size > 0:
            u = uppers[0]
        else:
            warnings.warn('n is too small to warrantee an exact CI other than the full range of the data.', RuntimeWarning)
            return np.array([-np.inf, x.max()]), [0, 1], [0, 1]
        quantiles = p_below_percentile[u]
        CI = np.array([-np.inf, np.sort(x)[u]])
        CI_prob = (u+1) / n
    elif alternative == 'greater':
        lows = np.where(p_below_percentile <= alpha/2)[0]
        if lows.size > 0:
            l = lows[-1]
        else:
            warnings.warn('n is too small to warrantee an exact CI other than the full range of the data.', RuntimeWarning)
            return np.array([x.min(), np.inf]), [0, 1], [0, 1]
        quantiles = p_below_percentile[l]
        CI = np.array([np.sort(x)[l], np.inf])
        CI_prob = l / n
    return CI, CI_prob, quantiles


def _exact_ci_percentile(x, p, alpha=0.05, alternative='two-sided', d_alpha=0.005):
    """
    DO NOT USE. This does not guarantee equal tailed. It is here as a reminder.
    
    Confidence 1 - alpha that the interval (Yi, Yj) contains the percentile p. Yi, Yj are order statistics.
    Attrs:
        - p:  percentile in (0, 1)
    https://online.stat.psu.edu/stat415/book/export/html/835
    """
    x_sorted = np.sort(x)
    n = x.size
    def confidence(low, high):    
        return np.array([math.comb(n, k) * p**k * (1-p)**(n-k) for k in range(low, high)]).sum()
    c = {}
    if alternative == 'two-sided':
        for low in range(1, n):
            for high in range(low+1, n+1):
                c[(low-1, high-1)] = confidence(low, high)
    elif alternative == 'less':
        for high in range(1, n+1):
            c[high-1] = confidence(0, high)
    elif alternative == 'greater':
        for low in range(1, n+1): 
            c[low-1] = 1 - confidence(0, low)
    c = pd.Series(c)
    c_pruned = c[(c - (1-alpha)).abs() < d_alpha]
    c_pruned = c_pruned.to_frame(name='confidence')
    if alternative == 'two-sided':
        low, high = [np.array(_) for _ in zip(*[x_sorted[list(i)] for i in c_pruned.index])]
        sample_stat = np.percentile(x, p*100)
        c_pruned['low'] = low
        c_pruned['high'] = high
        c_pruned['width'] = high - low
        c_pruned['asymmetry'] = (high - sample_stat) / (sample_stat - low)
        c_pruned = c_pruned.sort_values(['width', 'confidence'], ascending=[True, False])
    else:
        bound = [x_sorted[i] for i in c_pruned.index]
        if alternative == 'greater':
            c_pruned['low'] = bound 
            c_pruned['high'] = np.inf
        else:
            c_pruned['low'] = -1 * np.inf
            c_pruned['high'] = bound
    return c_pruned

def _find_ci_percentile_exact_confidence(ci, x, p, tol=1e-3):
    """
    DO NOT USE. This does not guarantee equal tailed CI.
    
    Given a CI 'ci' from data 'x' of a percentile statistic, return the exact confidence it provides.
    """
    if not math.isfinite(ci[0]):
        alternative = 'less'
        prune = lambda c: c[(c['high'] - ci[1]).abs() < tol]
    elif not math.isfinite(ci[1]):
        alternative = 'greater'
        prune = lambda c: c[(c['low'] - ci[0]).abs() < tol]
    else:
        alternative = 'two-sided'
        prune = lambda c: c[((c['low'] - ci[0]).abs() < tol) & ((c['high'] - ci[1]).abs() < tol)]
    c = _exact_ci_percentile(x, p, alternative=alternative, d_alpha=0.5)
    return prune(c)


"""
Helper funcs for plotting
"""
import numpy as np
from colour import Color
from PIL import ImageColor
from .. import _helper
        
def rows_cols(nrows, ncols):
    rows, cols = (np.vstack(np.divmod(range(nrows*ncols), ncols)) + 1)
    return rows, cols

def break_str(x, separator="<br>"):
    x_list = x.split(" ")
    mid = len(x_list) // 2
    return "{} {} {}".format(" ".join(x_list[:mid]), separator, " ".join(x_list[mid:]))

def plotly_default_colors(maxlen=None):
    colors = ['#1f77b4', '#ff7f0e',  '#2ca02c',  '#d62728',  '#9467bd',  '#8c564b',  '#e377c2',  '#7f7f7f',  '#bcbd22',  '#17becf']
    return _helper.sequence_or_stream(colors, maxlen=maxlen)

def plotly_colors(maxlen=None):
    cs = """aliceblue, antiquewhite, aqua, aquamarine, azure,
                beige, bisque, black, blanchedalmond, blue,
                blueviolet, brown, burlywood, cadetblue,
                chartreuse, chocolate, coral, cornflowerblue,
                cornsilk, crimson, cyan, darkblue, darkcyan,
                darkgoldenrod, darkgray, darkgrey, darkgreen,
                darkkhaki, darkmagenta, darkolivegreen, darkorange,
                darkorchid, darkred, darksalmon, darkseagreen,
                darkslateblue, darkslategray, darkslategrey,
                darkturquoise, darkviolet, deeppink, deepskyblue,
                dimgray, dimgrey, dodgerblue, firebrick,
                floralwhite, forestgreen, fuchsia, gainsboro,
                ghostwhite, gold, goldenrod, gray, grey, green,
                greenyellow, honeydew, hotpink, indianred, indigo,
                ivory, khaki, lavender, lavenderblush, lawngreen,
                lemonchiffon, lightblue, lightcoral, lightcyan,
                lightgoldenrodyellow, lightgray, lightgrey,
                lightgreen, lightpink, lightsalmon, lightseagreen,
                lightskyblue, lightslategray, lightslategrey,
                lightsteelblue, lightyellow, lime, limegreen,
                linen, magenta, maroon, mediumaquamarine,
                mediumblue, mediumorchid, mediumpurple,
                mediumseagreen, mediumslateblue, mediumspringgreen,
                mediumturquoise, mediumvioletred, midnightblue,
                mintcream, mistyrose, moccasin, navajowhite, navy,
                oldlace, olive, olivedrab, orange, orangered,
                orchid, palegoldenrod, palegreen, paleturquoise,
                palevioletred, papayawhip, peachpuff, peru, pink,
                plum, powderblue, purple, red, rosybrown,
                royalblue, saddlebrown, salmon, sandybrown,
                seagreen, seashell, sienna, silver, skyblue,
                slateblue, slategray, slategrey, snow, springgreen,
                steelblue, tan, teal, thistle, tomato, turquoise,
                violet, wheat, white, whitesmoke, yellow,
                yellowgreen"""
    li = cs.split(',')
    li = [l.replace('\n','') for l in li]
    li = [l.replace(' ','') for l in li]
    return _helper.sequence_or_stream(li, maxlen=maxlen)

def color_gradient(start, end, n, to_hex=True):
    colors = [*Color(start).range_to(Color(end), n)]
    if to_hex:
        return [c.hex for c in colors]
    else:
        return colors

def color_std(color, opacity=0.3):
    """Colors for the band y_mean +- y_std"""
    return "rgba" + str((*ImageColor.getrgb(color), opacity))


"""
Move/delete data
"""
import os
from pathlib import Path
import re

parent_dir =  'data'

def move_files(keyword, folder=None, parent_dir=parent_dir, verbose=1):
    """
    Moves all files in the parent directory starting with a keyword to a folder.
    
    Attributes:    
        - keyword: keyword to look for during file matching-
        - folder: Created folder to store all files matched by the keyword
        - parend_dir: Initial directory in which the files are contained.
    """
    folder = folder if folder is not None else keyword
    Path(os.path.join(parent_dir, folder)).mkdir(exist_ok=True, parents=True)
    processed = 0
    for path in os.listdir(parent_dir):
        if path == folder:
            continue
        else:
            if re.findall(key, path):
                os.rename(os.path.join(parent_dir, path), 
                          os.path.join(parent_dir, folder, path))
                processed += 1
    if verbose:
        print(f"Moved {processed} files from '{parent_dir}' to '{folder}' subdirectory.")        
    return

def delete_files_by_ext(parent_dir, extensions, verbose=1):
    """
    Removes files in parent_dir with extensions starting by (or being equal to) a certain character.
    """
    deleted = 0
    for path in os.listdir(parent_dir):
        for extension in extensions:
            if os.path.splitext(path)[1].startswith(extension):
                os.remove(os.path.join(parent_dir, path))
                deleted += 1
    if verbose:
        print(f"Deleted {deleted} files.")        
    return

def delete_stdin_files(parent_dir="nuredduna_programmes/stdin_files", verbose=1, completed_only=True, key="Done", key_search="any"):
    """
    Removes nuredduna standard input (stdin) files, of the form python.exxxx (s. error) and python.oxxxx (s. output).
    key_search:  sets how the key is looked in a file line:    - any:   key in any position
                                                               - start.
                                                               - end.
    """
    if completed_only:
        if key_search == "any":
            find_key = lambda l: key in l
        elif key_search == "start":
            find_key = lambda l: l.startswith(key)
        elif key_search == 'end':
            find_key = lambda l: l.endswith(key)
        else:
            raise ValueError(f"key_search '{key_search}' not valid. Available: 'any', 'start', 'end'.")
        
        deleted_programs = 0
        deleted_files = 0
        for f in os.listdir(parent_dir):
            if f.endswith("out"):
                ff = os.path.join(parent_dir, f)
                if any(find_key(l) for l in open(ff).readlines()):
                    os.remove(ff)
                    deleted_programs += 1
                    deleted_files += 1
                    ff_err = "{}.err".format(os.path.splitext(ff)[0])
                    if Path(ff_err).exists():
                        os.remove(ff_err)
                        deleted_files += 1
        if verbose:
            print(f"Deleted {deleted_files} files ({deleted_programs} completed programmes).")   
    else:
        delete_files_by_ext(parent_dir, [".e", ".o"], verbose)
    return

def empty_trash(verbose=1):
    home = os.path.expanduser("~")
    binDir = f"{home}/.local/share/Trash"
    deleted = 0
    for root, dirs, files in os.walk(binDir, topdown=False):
        for name in files:
            os.remove(os.path.join(root, name))
            deleted += 1
    if verbose:
        print(f"Deleted {deleted} files.")
    return


import inspect
import warnings
try:
    import rpy2.robjects as ro
    from rpy2.robjects import r, pandas2ri, numpy2ri
    from rpy2.robjects.packages import importr
    pandas2ri.activate()
    ro.numpy2ri.activate()
except:
    pass

def load_R_pkg(pkg):
    """Loads R package, installing it if needed."""
    if not ro.packages.isinstalled(pkg):
        warnings.warn(f"Package '{pkg}' not found. Proceeding to install it ...", RuntimeWarning)
        utils = importr("utils")
        utils.install_packages(pkg)
    r(f"library({pkg})")
    return

def attr_preprocess(func, attrs, clear_env=False):
    """
    Python enviroment -> R environment. 
    Returns: str version of **kwargs.
    """
    var_kwd_names = [p.name for p in inspect.signature(func).parameters.values() if p.kind.name == "VAR_KEYWORD"]
    if var_kwd_names:
        var_kwd_name = var_kwd_names[0]
        attrs.update(attrs[var_kwd_name])
        if len(attrs[var_kwd_name]) > 0:
            kwargs_str = ", " + ", ".join([f"{k}={k}" for k in attrs[var_kwd_name].keys()])
        else:
            kwargs_str = ""
        del attrs[var_kwd_name]
    
    if clear_env:
        ro.globalenv.clear()
    for k, v in attrs.items():
        if v is not None:
            ro.globalenv[k.replace("_", ".")] = v
    return kwargs_str


"""
Numba-compatible permutation tests, with/without pairing (including block pairing), for the difference and the ratio.

For statistics other than mean & median, pass the statistic function to 'get_permutation_test'.
"""
import types
import inspect
import numpy as np
from numba import njit

@njit
def _permutation_pvalue(alternative, perm_sample, stat_0, tolerance):
    R = perm_sample.shape[0]
    if alternative == "greater":
        return (1 + (perm_sample >= stat_0 - tolerance).sum()) / (R + 1)
    elif alternative == "less":
        return (1 + (perm_sample <= stat_0 + tolerance).sum()) / (R + 1)
    elif alternative == "two-sided":
        return min(2 * np.fmin((1 + (perm_sample >= stat_0 - tolerance).sum()) / (R + 1),
                               (1 + (perm_sample <= stat_0 + tolerance).sum()) / (R + 1)),
                   1)
    else:
        raise ValueError("alternative not valid. Available: 'two-sided', 'less', 'greater'.")
    
def _compile_func(func_string, env_global, use_numba):
    if func_string.startswith("    "): # Unexpected indentation.
        func_string = func_string[4:].replace("\n    ", "\n")
    code_obj = compile(func_string, '<string>', 'exec')
    if isinstance(code_obj.co_consts[-1], tuple):
        f = types.FunctionType(code_obj.co_consts[-4], env_global, argdefs=code_obj.co_consts[-1])
    else:
        f = types.FunctionType(code_obj.co_consts[-3], env_global, argdefs=code_obj.co_consts[:-3])
    if use_numba:
        return njit(f)
    else:
        return f

def _get_permutation_test_not_paired(stat_func, use_numba=True):
    """
    Generate a function for the permutation test using the statistic 'stat_func'. 
    This statistic can or not be numba compatible. Set 'use_numba' accordingly.
    """
    def permutation_test_STAT_FUNC_NAME(X1, X2, ratio=False, R=100000, alternative="two-sided", tolerance=1.5e-8, seed=0):
        """
        Permutation test for 'STAT_FUNC_NAME'. Permutations occur among all elements of X1 and X2.
        
        H0: X1 and X2 come from the same distribution F. The joint distribution would be independent on the variables: g(x, y) = f(x)f(y)
           
        Attrs:
            - alternative: one of 'two-sided', 'less', 'greater'.
                           Example of calculation for the mean mu:
                                                - greater:    H0 compatible: mu_1 / mu_2 < observed,       H1 compatible: mu_1 / mu_2  >= observed  
                              ratio             - less:       H0 compatible: mu_1 / mu_2 > observed,       H1 compatible: mu_1 / mu_2  <= observed  
                                                - two-sided:  H0 compatible: mu_1 / mu_2 = observed,       H1 compatible: mu_1 / mu_2  != observed 
            
                                                - greater:    H0 compatible: mu_1 - mu_2 < observed,       H1 compatible: mu_1 - mu_2  >= observed  
                              not ratio         - less:       H0 compatible: mu_1 - mu_2 > observed,       H1 compatible: mu_1 - mu_2  <= observed  
                                                - two-sided:  H0 compatible: mu_1 - mu_2 = observed,       H1 compatible: mu_1 - mu_2  != observed 
            - R:           number of resamples (permutations).
            - tolerance:   tolerance for numerical stability.
            - seed: seed   for random number generation.
        Returns: p-value
        """
        n1 = X1.size
        if ratio:
            def aux(X, n1): # pass n1 to avoid numba error.
                return stat_func(X[:n1]) / stat_func(X[n1:])
        else:
            def aux(X, n1):
                return stat_func(X[:n1]) - stat_func(X[n1:])
        X = np.hstack((X1, X2))
        stat_0 = aux(X, n1)
           
        perm_sample = np.empty((R)) # permutation distribution
        np.random.seed(seed)
        for i in range(R):       
            np.random.shuffle(X)
            perm_sample[i] = aux(X, n1)
        return _permutation_pvalue(alternative, perm_sample, stat_0, tolerance)
    
    func_source_edit = inspect.getsource(permutation_test_STAT_FUNC_NAME).replace("STAT_FUNC_NAME", stat_func.__name__)
    env = globals()
    env.update(locals())
    return _compile_func(func_source_edit, env, use_numba)
    
def _get_permutation_test_paired(stat_func, use_numba=True):
    """
    Generate a function for the paired permutation test using the statistic 'stat_func'. 
    This statistic can or not be numba compatible. Set 'use_numba' accordingly.
    """
    def permutation_test_paired_STAT_FUNC_NAME(X1, X2, ratio=False, R=100000, alternative="two-sided", tolerance=1.5e-8, seed=0):
        """
        Paired permutation test for 'STAT_FUNC_NAME'. Permutations occur only between each pair:   (x1, x2)  <->  (x2, x1).
        
        H0: (X1, X2) come from a symmetric joint distribution: p(x, y) = p(y, x).
        
        Attrs:
            - alternative: one of 'two-sided', 'less', 'greater'.
                           Example of calculation for the mean mu:
                                                - greater:    H0 compatible: mu_1 / mu_2 < observed,       H1 compatible: mu_1 / mu_2  >= observed  
                              ratio             - less:       H0 compatible: mu_1 / mu_2 > observed,       H1 compatible: mu_1 / mu_2  <= observed  
                                                - two-sided:  H0 compatible: mu_1 / mu_2 = observed,       H1 compatible: mu_1 / mu_2  != observed 
            
                                                - greater:    H0 compatible: mu_1 - mu_2 < observed,       H1 compatible: mu_1 - mu_2  >= observed  
                              not ratio         - less:       H0 compatible: mu_1 - mu_2 > observed,       H1 compatible: mu_1 - mu_2  <= observed  
                                                - two-sided:  H0 compatible: mu_1 - mu_2 = observed,       H1 compatible: mu_1 - mu_2  != observed 
            - R:           number of resamples (permutations).
            - tolerance:   tolerance for numerical stability.
            - seed: seed   for random number generation.
        Returns: p-value
        """
        if ratio:
            def aux(X_paired):
                return stat_func(X_paired[0]) / stat_func(X_paired[1])
        else:
            def aux(X_paired):
                return stat_func(X_paired[0]) - stat_func(X_paired[1])
            
        n = X1.size
        X_paired = np.vstack((X1, X2))
        stat_0 = aux(X_paired)
        
        perm_sample = np.empty((R))
        np.random.seed(seed)
        for i in range(R):
            shuffle = np.random.randint(0, 2, size=n) == 1
            X_shifted = X_paired[:, shuffle][::-1]
            X_still = X_paired[:, ~shuffle]
            X_perm = np.hstack((X_shifted, X_still))
            perm_sample[i] = aux(X_perm)
        return _permutation_pvalue(alternative, perm_sample, stat_0, tolerance)
    
    func_source_edit = inspect.getsource(permutation_test_paired_STAT_FUNC_NAME).replace("STAT_FUNC_NAME", stat_func.__name__)
    env = globals()
    env.update(locals())
    return _compile_func(func_source_edit, env, use_numba)
    
def _get_permutation_test_block(stat_func, use_numba=True):
    """
    Generate a function for the block permutation test using the statistic 'stat_func'.
    This statistic can or not be numba compatible. Set 'use_numba' accordingly.
    Permutations occur only between blocks: X1(block i) <-> X2(block i). 
    """
    def permutation_test_block_STAT_FUNC_NAME(X, Y, ratio=False, R=100000, alternative="two-sided", tolerance=1.5e-8, seed=0):
        """
        Block permutation test for 'STAT_FUNC_NAME'. Permutations occur only between blocks: X1(block i) <-> X2(block i). 
        
        H0: (X, Y) come from a joint distribution symmetric between each block:   p(x, y) = p(y, x)  <=> y, x belong to the same block.
        
        Attrs:
            - X, Y:        ragged arrays or tuples. Each element is an array containing the results for a block. 
            - alternative: one of 'two-sided', 'less', 'greater'.
                           Example of calculation for the mean mu:
                                                - greater:    H0 compatible: mu_1 / mu_2 < observed,       H1 compatible: mu_1 / mu_2  >= observed  
                              ratio             - less:       H0 compatible: mu_1 / mu_2 > observed,       H1 compatible: mu_1 / mu_2  <= observed  
                                                - two-sided:  H0 compatible: mu_1 / mu_2 = observed,       H1 compatible: mu_1 / mu_2  != observed 
            
                                                - greater:    H0 compatible: mu_1 - mu_2 < observed,       H1 compatible: mu_1 - mu_2  >= observed  
                              not ratio         - less:       H0 compatible: mu_1 - mu_2 > observed,       H1 compatible: mu_1 - mu_2  <= observed  
                                                - two-sided:  H0 compatible: mu_1 - mu_2 = observed,       H1 compatible: mu_1 - mu_2  != observed 
            - R:           number of resamples (permutations).
            - tolerance:   tolerance for numerical stability.
            - seed: seed   for random number generation.
        Returns: p-value
        """
        if ratio:
            def aux(X1, X2):
                return stat_func(X1) / stat_func(X2)
        else:
            def aux(X1, X2):
                return stat_func(X1) - stat_func(X2)
            
        def stack(arr_list):
            return np.array([a for arr in arr_list for a in arr])
            
        stat_0 = aux(stack(X), stack(Y))    
        perm_sample = np.empty((R))
        np.random.seed(seed)
        for i in range(R):
            Xi = []
            Yi = []
            for xi, yi in zip(X, Y):
                z = np.hstack((xi, yi))
                np.random.shuffle(z)
                Xi.append(z[:xi.size])
                Yi.append(z[xi.size:])
            perm_sample[i] = aux(stack(Xi), stack(Yi))
        return _permutation_pvalue(alternative, perm_sample, stat_0, tolerance)
    
    func_source_edit = inspect.getsource(permutation_test_block_STAT_FUNC_NAME).replace("STAT_FUNC_NAME", stat_func.__name__)
    env = globals()
    env.update(locals())
    return _compile_func(func_source_edit, env, use_numba)
    
def get_permutation_test(stat_func, pairing=None, use_numba=True):
    """
    Generate a function for the permutation test using the statistic 'stat_func'.
    This statistic can or not be numba compatible. Set 'use_numba' accordingly.
    
    pairing:
        - None:     Permutations occur among all elements of X1 and X2.
        - 'paired': Permutations occur only between each pair:   (x1, x2)  <->  (x2, x1).   (x1, x2) are elements of (X1, X2).
        - 'block':  Permutations occur only between blocks:      (x1_b, x2_b)  <->  (x2_b, x1_b).   x1_b in X1(block b), x2_b in X2(block b).
    
    Returns: function that computes the permutation test using the statistic 'stat_func'.
    """
    if pairing is None:
        return _get_permutation_test_not_paired(stat_func, use_numba)
    elif pairing == 'paired':
        return _get_permutation_test_paired(stat_func, use_numba)
    elif pairing == 'block':
        return _get_permutation_test_block(stat_func, use_numba)
    else:
        raise ValueError(f"pairing '{pairing}' is not valid. Available: None, 'paired', 'block'.")

permutation_test_mean = get_permutation_test(np.mean)
permutation_test_mean = get_permutation_test(np.median)
permutation_test_paired_mean = get_permutation_test(np.mean, pairing='paired')
permutation_test_paired_mean = get_permutation_test(np.median, pairing='paired')
permutation_test_block_mean = get_permutation_test(np.mean, pairing='block')
permutation_test_block_mean = get_permutation_test(np.median, pairing='block')


## Permutation test for mean/median of the difference or ratio
 
@njit
def permutation_test_paired_diffmean(X1, X2, R=int(1e5)-1, alternative="two-sided", tolerance=1.5e-8, seed=0):
    """
    Paired permutation test optimized for the mean of the differences. about 2x faster than the default _permutation_test_2sample_paired with stat_func = np.mean.
    Coincides with the test for the differences in means. (Not as the median)
    Attrs:
            alternative:  - greater:    H0: mu(X1-X2) < <X1-X2>_sample,       H1: mu(X1-X2) >= <X1-X2>_sample
                          - less:       H0: mu(X1-X2) > <X1-X2>_sample,       H1: mu(X1-X2) <= <X1-X2>_sample
                          - two-sided:  H0: mu(X1-X2) = <X1-X2>_sample,       H1: mu(X1-X2) != <X1-X2>_sample
    Returns: p-value
    """
    dX = X1 - X2
    n = dX.size
    stat_0 = dX.mean()
    
    perm_sample = np.empty((R))
    np.random.seed(seed)
    for i in range(R):        
        shuffle = np.random.randint(0, 2, size=n) == 1
        dX_perm = dX.copy()
        dX_perm[shuffle] *= -1
        perm_sample[i] = (dX_perm).mean()
    
    return _permutation_pvalue(alternative, perm_sample, stat_0, tolerance)
        
@njit
def permutation_test_paired_diffmedian(X1, X2, R=int(1e5)-1, alternative="two-sided", tolerance=1.5e-8, seed=0):
    """
    Paired permutation test optimized for the median of the differences. about 2x faster than the default _permutation_test_2sample_paired with stat_func = np.mean.
    NOTE: Does not coincide with the test for the differences in medians.
    
    Attrs:
            alternative:  - greater:    H0: me(X1-X2) < observed,       H1: me(X1-X2) >= observed
                          - less:       H0: me(X1-X2) > observed,       H1: me(X1-X2) <= observed
                          - two-sided:  H0: me(X1-X2) = observed,       H1: me(X1-X2) != observed.
    Returns: p-value
    """
    dX = X1 - X2
    n = dX.size
    stat_0 = np.median(dX)
    
    perm_sample = np.empty((R))
    np.random.seed(seed)
    for i in range(R):        
        shuffle = np.random.randint(0, 2, size=n) == 1
        dX_perm = dX.copy()
        dX_perm[shuffle] *= -1
        perm_sample[i] = np.median(dX_perm)
    
    return _permutation_pvalue(alternative, perm_sample, stat_0, tolerance)

@njit
def permutation_test_paired_ratio_median(X1, X2, R=int(1e5)-1, alternative="two-sided", tolerance=1.5e-8, seed=0):
    """
    Paired permutation test for the median ratio.
    
    Attrs:
            alternative: 
                                           - greater:    H0: me_1 / me_2 < observed,       H1: me_1 / me_2  >= observed  
                                           - less:       H0: me_1 / me_2 > observed,       H1: me_1 / me_2  <= observed  
                                           - two-sided:  H0: me_1 / me_2 = observed,       H1: me_1 / me_2  != observed 
            
    Returns: p-value
    """
    def aux(X_paired):
            return np.median(X_paired[0] / X_paired[1])
        
    n = X1.size
    X_paired = np.vstack((X1, X2))
    stat_0 = aux(X_paired)
    
    perm_sample = np.empty((R))
    np.random.seed(seed)
    for i in range(R):
        shuffle = np.random.randint(0, 2, size=n) == 1
        X_shifted = X_paired[:, shuffle][::-1]
        X_still = X_paired[:, ~shuffle]
        X_perm = np.hstack((X_shifted, X_still))
        perm_sample[i] = aux(X_perm)
    
    return _permutation_pvalue(alternative, perm_sample, stat_0, tolerance)
    
@njit
def permutation_test_paired_ratio_mean(X1, X2, R=int(1e5)-1, alternative="two-sided", tolerance=1.5e-8, seed=0):
    """
    Paired permutation test for the median ratio.
    
    Attrs:
            alternative: 
                                           - greater:    H0: me_1 / me_2 < observed,       H1: me_1 / me_2  >= observed  
                                           - less:       H0: me_1 / me_2 > observed,       H1: me_1 / me_2  <= observed  
                                           - two-sided:  H0: me_1 / me_2 = observed,       H1: me_1 / me_2  != observed 
            
    Returns: p-value
    """
    def aux(X_paired):
            return np.mean(X_paired[0] / X_paired[1])
        
    n = X1.size
    X_paired = np.vstack((X1, X2))
    stat_0 = aux(X_paired)
    
    perm_sample = np.empty((R))
    np.random.seed(seed)
    for i in range(R):
        shuffle = np.random.randint(0, 2, size=n) == 1
        X_shifted = X_paired[:, shuffle][::-1]
        X_still = X_paired[:, ~shuffle]
        X_perm = np.hstack((X_shifted, X_still))
        perm_sample[i] = aux(X_perm)
    
    return _permutation_pvalue(alternative, perm_sample, stat_0, tolerance)



"""
Explore PCA results. In the future will contain more algorithms.
Includes:
    - Variance visualization
    - Importance of each feature (how much each feature contributes to the PCA variance explained)
    - PCA components visualization in 2D.
"""
try:
    from sklearn.decomposition import PCA
    from sklearn.preprocessing import StandardScaler
except:
    pass
try:
    import plotly.graph_objects as go
    from .plots.plotly_utils import get_figure
except:
    pass
    
def pca_fit(X):
    x = StandardScaler().fit_transform(X)    
    pca = PCA()
    pca.fit(x)
    return pca

def pca_variance(X, cummulative=True):
    """Variance explained by each PCA component."""
    pca = pca_fit(X)
    y = pca.explained_variance_ratio_.cumsum() if cummulative else pca.explained_variance_ratio_
    fig = get_figure(yaxis_title_text="Variance explained", yaxis_range=[0,1], xaxis_title_text="PCA component")
    fig.add_trace(go.Bar(x=np.arange(len(pca.components_))+1, y=y, showlegend=False))
    if cummulative:
        fig.add_hline(y=0.9, line=dict(color='orange', width=5, dash='dot'))
        fig.add_hline(y=0.95, line=dict(color='red', width=5, dash='dot'))
    return fig

def pca_feature_importance(df, xaxis_tickfont_size=20, yaxis_range=None):
    """PCA importance of feature Z_i = sum(|PCA projections of Z_i| * explained_variance of each projection)."""
    pca = pca_fit(df.values)
    col_importance = []
    for pca_comp in pca.components_.T:
        col_importance.append(np.sum(np.abs(pca_comp) * pca.explained_variance_ratio_))
    col_importance = np.array(col_importance)
    # Alternatively: col_importance = (np.abs(pca.components_) * pca.explained_variance_ratio_[:, None]).sum(axis=0)
    col_importance /= col_importance.sum()
    fig = get_figure(yaxis_title_text="PCA importance", yaxis_range=yaxis_range, xaxis_title_text="Feature", xaxis_tickfont_size=xaxis_tickfont_size)
    fig.add_trace(go.Bar(x=df.columns, y=col_importance, showlegend=False))
    return fig

def visualize_pca_components(df, features, dx=[-15, 10], dy=[-10, 10], **fig_kwargs):
    """Visualize pca components projections in 2D. dx, dy are offset values for the x and y axes."""
    pca = pca_fit(df.values)
    feature_map = {f: i for i, f in enumerate(df.columns)}
    loadings = pca.components_.T * np.sqrt(pca.explained_variance_)
    fig = get_figure(xaxis_title="PCA 1", yaxis_title="PCA 2", **fig_kwargs)
    for feature in features:
        i = feature_map[feature]
        fig.add_shape(type='line', x0=0, y0=0, x1=loadings[i, 0], y1=loadings[i, 1])
        fig.add_annotation(x=loadings[i, 0], 
                           y=loadings[i, 1] if loadings[i,1] > 0 else loadings[i,1] - 4,
                           ax=0, ay=0, xanchor="center", yanchor="bottom", text=feature,
        )
    fig.update_layout(xaxis=dict(range=[dx[0] + loadings.min(axis=0)[0], dx[1]+loadings.max(axis=0)[0]], 
                             showgrid=False, zeroline=False, visible=True, tickvals=[]),
                  yaxis=dict(range=[dy[0] + loadings.min(axis=0)[1], dy[1]+loadings.max(axis=0)[1]],
                             showgrid=False, zeroline=False, visible=True, tickvals=[]),
                  paper_bgcolor='rgba(0,0,0,0)', plot_bgcolor='rgba(0,0,0,0)'
                 )
    return fig



"""
Helper funcs for plotly figures
"""
import numpy as np
import pandas as pd
import re
import warnings
try:
    import plotly.express as px
    import plotly.graph_objects as go
    from plotly.subplots import make_subplots
except:
    warnings.warn("'plotly' not available.", RuntimeWarning)
try:
    import colorlover as cl
except:
    warnings.warn("'colorlover' not available.", RuntimeWarning)
from collections import defaultdict
from collections.abc import Iterable
from functools import partial
from .. import _helper
from .base import color_std, plotly_default_colors

def add_offset(x0, xf, offset=0.05):
    """x0 (xf) == lower (upper) limit for the axis range."""
    inverse_transform = lambda *xs: [(xf-x0)*x + x0 for x in xs]
    return inverse_transform(-offset, 1+offset)
    
def get_common_range(fig, axes=["x", "y"], offset=[0.05, 0.05]):
    data = defaultdict(list)
    for plot in fig.data:
        for ax in axes:          
            if hasattr(plot, f"error_{ax}") and getattr(plot, f"error_{ax}").array is not None:
                additions = [np.array([*plot[f"error_{ax}"]["array"]]), -np.array([*plot[f"error_{ax}"]["array"]])] 
            else:
                additions = [0]
            for addition in additions:
                try:
                    arr = (plot[ax] + addition)[~np.isnan(plot[ax])]
                except:
                    continue
                arr_min, arr_max = arr.min(), arr.max()
                data[f"{ax}-min"].append(arr_min)
                data[f"{ax}-max"].append(arr_max)
    for k, v in data.items():
        func = min if "min" in k else max
        data[k] = func(v)
    ranges = {ax: add_offset(data[f"{ax}-min"], data[f"{ax}-max"], offset=off) for ax, off in zip(axes, offset)}
    return ranges

def get_nplots(fig):
    return sum(1 for x in fig.layout if "xaxis" in x)

def mod_delete_axes(fig, axes=["x", "y"]):
    non_visible_axes_specs = dict(visible=False, showgrid=False, zeroline=False) 
    return {f"{ax}axis{i}": non_visible_axes_specs for ax in axes for i in [""] + [*range(1, get_nplots(fig) + 1)]}

def get_mod_layout(key, val=None):
    def mod_layout(fig, val, axes=["x","y"]):
        if isinstance(val, Iterable) and not isinstance(val, str):
            return {"{}axis{}_{}".format(ax, i, key): v for (ax, v) in zip(axes, val) for i in [""] + [*range(1, get_nplots(fig) + 1)]}
        else:
            return {"{}axis{}_{}".format(ax, i, key): val for ax in axes for i in [""] + [*range(1, get_nplots(fig) + 1)]}
    if val is None:
        return mod_layout
    else:
        def mod_layout_fixed_val(fig, axes=["x", "y"]):
            return mod_layout(fig, val, axes)
        return mod_layout_fixed_val

mod_dashes           = partial(_helper.sequence_or_stream, ["solid", "dash", "dot"])
mod_ticksize         = get_mod_layout("tickfont_size")
mod_logaxes          = get_mod_layout("type", "log") 
mod_expfmt           = get_mod_layout("exponentformat", "power")
mod_range            = get_mod_layout("range")
mod_logaxes_expfmt   = lambda fig, axes=["x", "y"]: {**mod_logaxes(fig, axes=axes), **mod_expfmt(fig, axes=axes)}

def mod_common_range(fig, axes=["x", "y"], **kwargs):
    return mod_range(fig, val=get_common_range(fig, axes=axes, **kwargs), axes=axes)

def get_figure(height=800, width=1000, ticksize=32, font_size=40, margin=None, font_family="sans-serif", hovermode=False, delete_axes=False, **kwargs):
    fig = go.Figure(layout=dict(margin=dict(l=100, r=20, b=80, t=20, pad=1) if margin is None else margin,
                                height=height, width=width, yaxis=dict(tickfont_size=ticksize),
                                xaxis=dict(tickfont_size=ticksize), font_size=font_size, legend_font_size=font_size,
                                font_family=font_family, hovermode=hovermode,
                                **kwargs))
    if delete_axes:
        fig.update_layout(**mod_delete_axes(fig), margin=dict(l=0, t=0, b=0, r=0), paper_bgcolor="rgba(0,0,0,0)", plot_bgcolor="rgba(0,0,0,0)")
    return fig

def get_subplots(cols, rows=1, horizontal_spacing=0.03, vertical_spacing=0.03, height=None, width=2500, ticksize=32, font_size=40, font_family="sans-serif",
                 hovermode=False, delete_axes=False, shared_xaxes=True, shared_yaxes=True, layout_kwargs={}, 
                 **make_subplots_kwargs):
    height = 800*rows if height is None else height
    fig = make_subplots(figure=go.Figure(layout=dict(margin=dict(l=100, r=20, b=80, t=60, pad=1), height=height, width=width)),
                        shared_yaxes=shared_yaxes, shared_xaxes=shared_xaxes,                        
                        horizontal_spacing=horizontal_spacing, vertical_spacing=vertical_spacing, rows=rows, cols=cols,
                        **make_subplots_kwargs
                       )
                    
    fig.for_each_annotation(lambda a: a.update(font={'size':font_size, "family":font_family}))
    fig.update_layout(**mod_ticksize(fig, val=ticksize), legend_font_size=font_size, hovermode=hovermode, **layout_kwargs)
    if delete_axes:
        fig.update_layout(**mod_delete_axes(fig), margin=dict(l=0, t=0, b=0, r=0), paper_bgcolor="rgba(0,0,0,0)", plot_bgcolor="rgba(0,0,0,0)")
    return fig

def transparent_colorscale(fig, threshold=1e-10):
    """Values below threshold are invisible."""
    colorscale = fig.layout["coloraxis"]["colorscale"]
    low_limit = colorscale[0]
    new_low_limit = (threshold, low_limit[1])
    new_colorscale = ((0, 'rgba(0,0,0,0)'), new_low_limit, *colorscale[1:])
    return new_colorscale

def multiindex_to_label(i, depth=2):
    return [i.get_level_values(k).to_list() for k in range(depth)]

def set_multicategory_from_df(fig, df):
    fig.update_layout(xaxis_type="multicategory", yaxis_type="multicategory")
    fig.data[0]["x"] = multiindex_to_label(df.columns)
    fig.data[0]["y"] = multiindex_to_label(df.index)
    return

def CI_plot(x, y, CI, label=None, width=0.05, ms=10, color='rgba(255, 127, 14, 0.3)', fig=None, x_title=None, y_title=None):
    """
    Box plot where the box corresponds to the CI.
    
    Attributes:
        - x:    x coordinate for the CI
        - y:    value of the magnitude for the sample. Example: the mean if CI is a CI for the mean.
        - CI:   Confidence interval for y.
    """
    if fig is None:
        fig = get_figure(xaxis_title=x_title, yaxis_title=y_title)
    for i, (ci, x_val, ci_stat) in enumerate(zip(CI, x, y)):
        fig.add_trace(go.Scatter(x=[x_val]*2, y=ci[::-1], showlegend=False, mode="markers",
                                 marker=dict(color=color, symbol=["arrow-bar-down", "arrow-bar-up"], size=ms, line=dict(color="gray", width=2))
                                ))
        fig.add_shape(type="rect", xref="x", yref="y", line=dict(color="gray",width=3), fillcolor=color, x0=i-width, y0=ci[0], x1=i+width, y1=ci[1])
        fig.add_shape(type="line", xref="x", yref="y", line=dict(color="gray", width=4),  x0=i-width, y0=ci_stat, x1=i+width, y1=ci_stat)
    if label is not None:
        yrange = [*get_common_range(fig, axes=["y"]).values()][0]
        fig.add_trace(go.Scatter(x=[1000], y=[1000], mode="markers", name=label, showlegend=True,
                                 marker=dict(symbol="square", color=color, size=22), line=dict(color="gray", width=2)))
        fig.update_layout(**mod_range(fig, ([-0.25, len(x)-0.75], yrange)))
    return fig

def permtest_plot(df, H1="", colorscale="Inferno", log=True, height=800, width=1000, font_size=40, bar_len=0.9, bar_x=0.95, bar_thickness=100):
    """H1 should not contain latex code. Use unicode and HTML for super/sub-indices."""
    if log:
        df = np.log10(df)
        zmin, zmax = np.log10(0.05), 0
        legtitle = "log<sub>10</sub>P-value"
    else:
        zmin, zmax = None, None
        legtitle = "P-value"
    fig = px.imshow(df, zmin=zmin, zmax=zmax, color_continuous_scale=colorscale)
    fig.update_layout(coloraxis_colorbar=dict(len=bar_len, x=bar_x, title=f"{legtitle}<br>H<sub>1</sub>: {H1}", thickness=bar_thickness),
                      height=height, width=width, font_size=font_size, hovermode=False,
                      margin=dict(l=0, b=0, t=0, r=0)
                     )
    return fig
    
def violin(df, CI=None, CI_line="mean", **CI_kwargs):
    """
    Violin plot including optionally the CI.
    
    Attributes:
        - df:   melted DataFrame. Contains only two columns: variable name (x) and value (y).
                                  The column names set the OX and OY labels.
    """
    x, y = df.columns
    fig = get_figure(xaxis_title=x, yaxis_title=y)
    fig.add_trace(go.Violin(x=df[x], y=df[y], showlegend=False))
    if CI is not None:
        fig = CI_plot(df[x].unique(), getattr(df.groupby(x), CI_line)().values.squeeze(), CI, fig=fig, **CI_kwargs)
    return fig

def rgb_to_hex(rgb):
    if isinstance(rgb, str):
        rgb = rgb_text_to_rgb(rgb)
    return '#' + '%02x%02x%02x' % rgb

def rgb_text_to_rgb(text):
    colors = re.findall(r'rgb\((.*?)\)', text)[0]
    return tuple(int(c) for c in colors.split(","))

def rgb_to_text(rgb):
    return 'rgb({},{},{})'.format(*rgb)

def p_value_colorscale(base_cs="RdBu_r", out='rgb-text', interpolate=True, step=0.05):
    """
    Colorscale where color difference is proportional to the probability of a type I error.
    Example: p=0.1 has double probability of type I error than p=0.05. 
             p=0.5 has 10 times the probability of type I error of p=0.05.
    """
    cs_len = int(0.5 / step)
    def keep_first_appearance(colorscale):
        s, c = zip(*colorscale)
        _, unique_idxs = np.unique(c, return_index=True)
        unique_idxs = np.sort(unique_idxs)
        unique_s = np.array(s)[unique_idxs]
        unique_c = np.array(c)[unique_idxs]
        return [[s, c] for s, c in zip(unique_s, unique_c)]
    
    if interpolate:
        cs_interpolated = cl.to_rgb(cl.interp([*zip(*px.colors.get_colorscale("RdBu_r"))][1], 2*cs_len + 1))
        cg_low = [rgb_to_hex(c) for c in cs_interpolated[:cs_len+1]]
        cg_upper = [rgb_to_hex(c) for c in cs_interpolated[cs_len+1:]][::-1]
        c0, cm, cf = [rgb_to_hex(c) for c in np.array(cs_interpolated)[[0, cs_len, -1]]]
    else:
        c0, cm, cf = [rgb_to_hex(c) for s, c in np.array(px.colors.get_colorscale(base_cs))[[0, 5, -1]]]
        cg_low = color_gradient(c0, cm, cs_len)
        cg_upper = color_gradient(cf, cm, cs_len)
    r = np.arange(step, 0.5+step, step).round(3)
    idxs = (cs_len - 1/(2*r)).astype(int)
    lower_half = keep_first_appearance([[0, c0]] + [[r_i, cg_low[i]] for i, r_i in zip(idxs, r)])
    upper_half = keep_first_appearance([[1, cf]] + [[r_i, cg_upper[i]] for i, r_i in zip(idxs, (1-r).round(3))])[::-1][1:]
    colorscale = lower_half + upper_half
    if out == 'rgb-text':
        func = lambda c: rgb_to_text(px.colors.hex_to_rgb(c))
    elif out == 'rgb':
        func = px.colors.hex_to_rgb
    elif out == 'hex':
        func = lambda c: c
    else:
        raise ValueError(f"out '{out}' not valid. Available: 'rgb-text', 'rgb', 'hex'.")
    colorscale = [[s, func(c)] for s, c in colorscale]
    return colorscale

def plot_cs(cs):
    from IPython.display import HTML
    if isinstance(cs[0], list):
        _, c = zip(*cs)
    else:
        c = cs
    return HTML(cl.to_html(cl.to_hsl(c)))    


"""
Numba version of bias-corrected and accelerated (BCa) bootstrap.
                 studentized.
                 studentized (variance-stabilized).
                 percentile.
"""
import numpy as np
import pandas as pd
from numba import njit, boolean
from itertools import product
from collections import defaultdict
import warnings
try:
    from scipy.special import ndtri, ndtr
except:
    warnings.warn('scipy not available. Numba BCa bootstrap will not work.', RuntimeWarning)
try:
    from statsmodels.nonparametric.smoothers_lowess import lowess
    from scipy.interpolate import interp1d # for interpolation of new data points
except:
    warnings.warn('scipy or statsmodels not available. Studentized bootstrap will not work.', RuntimeWarning)
try:
    shell = get_ipython().__class__.__name__
    if shell == 'ZMQInteractiveShell': # script being run in Jupyter notebook
        from tqdm.notebook import tqdm
    elif shell == 'TerminalInteractiveShell': #script being run in iPython terminal
        from tqdm import tqdm
except NameError:
    from tqdm import tqdm # Probably runing on standard python terminal.

from ..np_utils import numpy_fill 
from ._integration import simpson3oct_vec
from . import conf_interval
    
@njit
def resample_paired_nb(X, Y, func, output_len=1, R=int(1e5), seed=0):
    np.random.seed(seed)
    N = X.size    
    data_paired = np.vstack((X, Y)).T
    idxs_resampling = np.random.randint(low=0, high=N, size=R*N)
    data_resampled = data_paired[idxs_resampling].reshape(R, N, 2)    
    stat = func(X, Y)
    
    boot_sample = np.empty((R, output_len))
    for i, r in enumerate(data_resampled):
        x, y = r.T
        boot_sample[i] = func(x, y)
    return boot_sample

@njit
def resample_nb_X(X, R=int(1e5), seed=0, smooth=False, N=0):
    """X: array of shape (N_samples, n_vars)."""
    np.random.seed(seed)
    n, output_len = X.shape
    if N == 0:
        N = n
    idxs_resampling = np.random.randint(low=0, high=n, size=R*N)
    data_resampled = X[idxs_resampling].reshape(R, N, output_len)
    if smooth:
        def x_in_percentile(x):
            low, high  = np.percentile(x, [5, 95])
            z = x[(x>low) & (x<high)]
            return z
        def std_percentile(x):
            z = x_in_percentile(x)
            return z.std() / np.sqrt(z.size)
        h = np.array([std_percentile(x) for x in X.T])
        n_trimmed = x_in_percentile(X.T[0]).size
        for k, h_k in enumerate(h):
            data_resampled[:,:,k] += h_k * np.random.standard_t(n_trimmed, R*N)
    return data_resampled

@njit
def resample_nb(X, func, output_len=1, R=int(1e5), seed=0, smooth=False, N=0):
    """X: array of shape (N_samples, n_vars)."""
    data_resampled = resample_nb_X(X, R=R, seed=seed, smooth=smooth, N=N)
    
    boot_sample = np.empty((R, output_len))
    for i, r in enumerate(data_resampled):
        boot_sample[i] = func(r)
    return boot_sample

@njit
def resample_block_nb(X, Y, func, output_len=1, R=int(1e5), seed=0):
    """
    X, Y:   ragged arrays or tuples. Each element is an array containing the data for a block. 
    func:   numba function f: X,Y  ->  Z,   Z: 1D array of size output_len.
    """
    np.random.seed(seed)
    def stack(arr_list):
        return np.array([a for arr in arr_list for a in arr])
    
    n_x = [len(x) for x in X]
    n_y = [len(y) for y in Y]
    idxs_resampling_x = [np.random.randint(low=0, high=n, size=R*n) for n in n_x]
    idxs_resampling_y = [np.random.randint(low=0, high=n, size=R*n) for n in n_y]
    X_resampled = [x[idxs_resampling].reshape(R, n) for x, n, idxs_resampling in zip(X, n_x, idxs_resampling_x)]
    Y_resampled = [y[idxs_resampling].reshape(R, n) for y, n, idxs_resampling in zip(Y, n_y, idxs_resampling_y)]
    
    boot_sample = np.empty((R, output_len))
    for i in range(R):
        Xi = stack([x[i] for x in X_resampled])
        Yi = stack([y[i] for y in Y_resampled])
        boot_sample[i] = func(Xi, Yi)
    return boot_sample

def resample(X, func, output_len=1, R=int(1e4), seed=0):
    """X: array of shape (N_samples, n_vars)."""
    np.random.seed(seed)
    N = X.shape[0]
    idxs_resampling = np.random.randint(low=0, high=N, size=R*N)
    data_resampled = X[idxs_resampling].reshape(R, N, X.shape[1])
    
    boot_sample = np.empty((R, output_len))
    for i, r in enumerate(data_resampled):
        boot_sample[i] = func(r)
    return boot_sample


def resample_block(X, Y, func, output_len=1, R=int(1e5), seed=0):
    """
    X, Y:   ragged arrays or tuples. Each element is an array containing the data for a block. 
    func:   numba function f: X,Y  ->  Z,   Z: 1D array of size output_len.
    """
    np.random.seed(seed)
    
    n_x = [len(x) for x in X]
    n_y = [len(y) for y in Y]
    idxs_resampling_x = [np.random.randint(low=0, high=n, size=R*n) for n in n_x]
    idxs_resampling_y = [np.random.randint(low=0, high=n, size=R*n) for n in n_y]
    X_resampled = [x[idxs_resampling].reshape(R, n) for x, n, idxs_resampling in zip(X, n_x, idxs_resampling_x)]
    Y_resampled = [y[idxs_resampling].reshape(R, n) for y, n, idxs_resampling in zip(Y, n_y, idxs_resampling_y)]
    
    boot_sample = np.empty((R, output_len))
    for i in range(R):
        Xi = np.hstack([x[i] for x in X_resampled])
        Yi = np.hstack([y[i] for y in Y_resampled])
        boot_sample[i] = func(Xi, Yi)
    return boot_sample

    
@njit
def jackknife_resampling(data):
    """Performs jackknife resampling on numpy arrays.

    Jackknife resampling is a technique to generate 'n' deterministic samples
    of size 'n-1' from a measured sample of size 'n'. The i-th
    sample  is generated by removing the i-th measurement
    of the original sample.
    """
    n = data.shape[0]
    if data.ndim > 1:
        resamples = np.empty((n, n - 1) + data.shape[1:])
        base_mask = np.ones((n), dtype=boolean)
        for i in range(n): # np.delete does not support 'axis' argument in numba.
            mask_i = base_mask.copy()
            mask_i[i] = False
            resamples[i] = data[mask_i]
    else:
        resamples = np.empty((n, n - 1))
        for i in range(n):
            resamples[i] = np.delete(data, i)
    return resamples

@njit
def jackknife_stat_nb(data, statistic):
    resamples = jackknife_resampling(data)
    stats = np.array([statistic(r) for r in resamples])
    return stats

def jackknife_stat_two_samples(data, data2, statistic):
    jk_X = jackknife_resampling(data)
    jk_Y = jackknife_resampling(data2)
    jk_XY = [*product(jk_X, jk_Y)]
    stats = np.array([statistic(*r) for r in jk_XY])
    return stats

def jackknife_stat_(data, statistic):
    resamples = jackknife_resampling(data)
    stats = np.array([statistic(r) for r in resamples])
    return stats

def _percentile_of_score(a, score, axis, account_equal=False):
    """Vectorized, simplified `scipy.stats.percentileofscore`.

    Unlike `stats.percentileofscore`, the percentile returned is a fraction
    in [0, 1].
    """
    B = a.shape[axis]
    if account_equal:
        return ((a < score).sum(axis=axis) + (a <= score).sum(axis=axis)) / (2 * B)
    else:
        return (a < score).sum(axis=axis) / B
    
def CI_bca(data, statistic, data2=None, alternative='two-sided', alpha=0.05, R=int(1e5), account_equal=False, use_numba=True, n_min=5, **kwargs):
    """If data2 is provided, assumes a block resampling and statistic takes two arguments."""
    if alternative == 'two-sided':
        probs = np.array([alpha/2, 1 - alpha/2])
    elif alternative == 'less':
        probs = np.array([0, 1-alpha])
    elif alternative == 'greater':
        probs = np.array([alpha, 1])
    else:
        raise ValueError(f"alternative '{alternative}' not valid. Available: 'two-sided', 'less', 'greater'.")
    
    if data2 is None:
        N = data.shape[0]
        if N < n_min:
            warnings.warn(f"N={N} < n_min={n_min}. Avoiding computation (returning NaNs) ...")
            return np.array([np.NaN, np.NaN])
        else:
            resample_func = resample_nb if use_numba else resample
            theta_hat_b = resample_func(data[:,None] if data.ndim == 1 else data,
                                        statistic, R=R, **kwargs).squeeze()
    else:
        total_N = lambda data: np.sum([d.shape[0] for d in data])
        N = min([total_N(data), total_N(data2)])
        if N < n_min:
            warnings.warn(f"N={N} < n_min={n_min}. Avoiding computation (returning NaNs) ...")
            return np.array([np.NaN, np.NaN])
        else:
            resample_func = resample_block_nb if use_numba else resample_block
            theta_hat_b = resample_func(data, data2, statistic, R=R, **kwargs).squeeze()
            data = np.hstack(data)
            data2 = np.hstack(data2)
        
    alpha_bca = _bca_interval(data, data2, statistic, probs, theta_hat_b, account_equal, use_numba)[0]
    
    if np.isnan(alpha_bca).all(): 
        warnings.warn('CI shows there is only one value. Check data.', RuntimeWarning)
        if data2 is None:
            sample_stat = statistic(data)
        else:
            sample_stat = statistic(data, data2)
        return np.array([sample_stat, sample_stat])
    elif alternative == 'two-sided':
        return  np.percentile(theta_hat_b, alpha_bca*100, axis=0)
    elif alternative == 'less':
         return np.array([-np.inf, np.percentile(theta_hat_b, alpha_bca[0]*100, axis=0)])
    elif alternative == 'greater':
        return np.array([np.percentile(theta_hat_b, alpha_bca[0]*100, axis=0), np.inf])
    
def _bca_interval(data, data2, statistic, probs, theta_hat_b, account_equal, use_numba):
    """Bias-corrected and accelerated interval."""
    # calculate z0_hat
    if data2 is None:
        theta_hat = statistic(data)
    else:
        theta_hat = statistic(data, data2)
    percentile = _percentile_of_score(theta_hat_b, theta_hat, axis=-1, account_equal=account_equal)
    z0_hat = ndtri(percentile)

    # calculate a_hat
    if data2 is None:
        jackknife_computer = jackknife_stat_nb if use_numba else jackknife_stat
        theta_hat_jk = jackknife_computer(data, statistic)  # jackknife resample
    else:
        theta_hat_jk = jackknife_stat_two_samples(data, data2, statistic)
    n = theta_hat_jk.shape[0]
    theta_hat_jk_dot = theta_hat_jk.mean(axis=0) 

    U = (n - 1) * (theta_hat_jk_dot - theta_hat_jk)
    num = (U**3).sum(axis=0) / n**3
    den = (U**2).sum(axis=0) / n**2
    a_hat = 1/6 * num / (den**(3/2))

    # calculate alpha_1, alpha_2
    def compute_alpha(p):
        z_alpha = ndtri(p)
        num = z0_hat + z_alpha
        return ndtr(z0_hat + num/(1 - a_hat*num))
    alpha_bca = compute_alpha(probs[(probs != 0) & (probs != 1)])
    if (alpha_bca > 1).any() or (alpha_bca < 0).any():
        warnings.warn('percentiles must be in [0, 1]. bca percentiles: {}\nForcing percentiles in [0,1]...'.format(alpha_bca), RuntimeWarning)
        alpha_bca = np.clip(alpha_bca, 0, 1)
    return alpha_bca, a_hat  # return a_hat for testing

def vs_transform(data, bootstrap_estimates, se_bootstrap, precision=1e-3, frac=2/3):
    """
    Variance-stabilizing transformation.
    """
    n_stats = bootstrap_estimates.shape[1]
    g = np.empty((data.shape[0], n_stats))
    lowess_linear_interp = []
    for i, (b, se, d) in enumerate(zip(bootstrap_estimates.T,  se_bootstrap.T, data.T)):
        x, y = lowess(se, b, frac=frac).T
        f_linear = interp1d(np.unique(x), y=np.unique(y), bounds_error=False, kind='linear', fill_value='extrapolate')
        z_min = d.min()
        for k, z in enumerate(d):
            g[k, i] = simpson3oct_vec(vs_integrand, z_min, z, precision, f_linear)[0]
        lowess_linear_interp.append(f_linear)
    return g, lowess_linear_interp

def invert_CI(CI, z, g, lowess_linear_interp, frac=1/10, min_n=100, integration_precision=1e-4):
    CIs = np.empty(CI.shape)
    for k, (ci, zi, gi, f_linear) in enumerate(zip(CI, z.T, g.T, lowess_linear_interp)):
        n = zi.size
        if n < min_n:
            z_std = zi.std() / np.sqrt(zi.size)
            z_min = zi.min()
            extra_z = np.unique(np.linspace(z_min - z_std/2, z.max()+z_std/2, min_n - n))
            extra_g = np.empty((extra_z.size))
            for j, extra_zi in enumerate(extra_z):
                extra_g[j] = simpson3oct_vec(vs_integrand, z_min, extra_zi, integration_precision, f_linear)[0]
            zi = np.hstack((zi, extra_z))
            gi = np.hstack((gi, extra_g))
        g_l, z_l = lowess(zi, gi, frac=frac).T
        f_inv = interp1d(np.unique(g_l), y=np.unique(z_l), bounds_error=False, kind='linear', fill_value='extrapolate')
        finite_ci = np.isfinite(ci) # account for single-tail.
        CIs[k, finite_ci] = f_inv(ci[finite_ci])
        CIs[k, ~finite_ci] = ci[~finite_ci]
    return CIs

def compute_CI_studentized(base, results, studentized_results, alpha=0.05, alternative='two-sided'):
    R, output_len = results.shape
    bootstrap_estimate = results.mean(axis=0)
    errors = results - bootstrap_estimate
    std_err = np.asarray(np.sqrt(np.diag(errors.T.dot(errors) / R)))
    
    lower = np.empty((output_len))
    upper = np.empty((output_len))
    alpha_tails = alpha / 2 if alternative == 'two-sided' else alpha
    percentiles = 100 * np.array([[alpha_tails, 1.0 - alpha_tails]] * output_len)
    for i in range(output_len):
        lower[i], upper[i] = np.percentile(studentized_results[:, i], percentiles[i])
    if alternative == 'less':
        upper.fill(np.inf)
    elif alternative == 'greater':
        lower.fill(-1 * np.inf)
        
    # Basic and studentized use the lower empirical quantile to compute upper and vice versa.      
    lower_copy = lower + 0.0
    lower = base - upper * std_err
    upper = base - lower_copy * std_err
    CI = np.vstack((lower, upper)).T
    return CI

def vs_integrand(x, f_linear):
    """Integrand of the variance-stabilizing transformation."""
    clipped_f = np.clip(f_linear(x), 1e-8, None)
    if np.isnan(clipped_f).any():
        clipped_f = numpy_fill(clipped_f)
    return 1 / clipped_f

def cov(results, base=None, recenter=False):
    """
    reps : Number of bootstrap replications
    recenter : Whether to center the bootstrap variance estimator on the average of the bootstrap samples (True), or 
                       to center on the original sample estimate (False).
    """
    if recenter:
        errors = results - np.mean(results, 0)
    else:
        assert base is not None
        errors = results - base
    return errors.T.dot(errors) / results.shape[0]

def _bootstrap_studentized_resampling(data, statistic, alpha=0.05, R=10000, studentized_reps=100, recenter=False, se_func=None, seed=0, divide_by_se=True, smooth=False):
    base = np.asarray(statistic(data))
    output_len = base.size
    studentized_results = np.empty((R, output_len))
    results = np.empty((R, output_len))
    se_bootstrap = np.empty((R, output_len))
    n = data.shape[0]
    if divide_by_se:
        def get_studentized(data_r, result, seed):
            nested_resampling = resample_nb(data_r, statistic, R=studentized_reps, output_len=output_len, seed=seed, smooth=False)        
            std_err = np.sqrt(np.diag(cov(nested_resampling, result, recenter=recenter)))
            err = result - base
            t_result = err /std_err
            return t_result, std_err
    else:
        def get_studentized(data_r, result, seed):
            return result - base, np.NaN
    
    data_r = resample_nb_X(data, R=R, seed=seed, smooth=smooth)
    if se_func is None:
        for i, d_r in enumerate(data_r):
            result = statistic(d_r)
            t_result, std_err = get_studentized(d_r, result, i)
            results[i] = result
            studentized_results[i] = t_result # t = (x^ - x) / s
            se_bootstrap[i] = std_err
    else:
        for i, d_r in enumerate(data_r):
            result = statistic(d_r)
            se = se_func(d_r)
            results[i] = result
            studentized_results[i] = (result - base) / se
            se_bootstrap[i] = se
    return base, results, studentized_results, se_bootstrap

def CI_studentized(data, statistic, R=int(1e5), alpha=0.05, alternative='two-sided', smooth=False, vs=False, 
                   frac_g=2/3, frac_invert=1/10, studentized_reps=100, 
                   integration_precision=1e-4, **kwargs):
    assert alternative in ['two-sided', 'less', 'greater'], f"alternative '{alternative}' not valid. Available: 'two-sided', 'less', 'greater'."
    base, results, studentized_results, se_bootstrap = _bootstrap_studentized_resampling(data, statistic, smooth=smooth, R=R, studentized_reps=studentized_reps, **kwargs)
    if vs:
        g, lowess_linear_interp = vs_transform(data, results, se_bootstrap, precision=integration_precision, frac=frac_g)
        base_g, results_g, studentized_results_g, _ = _bootstrap_studentized_resampling(g, statistic, R=R, divide_by_se=False, smooth=False)
        CI = invert_CI(compute_CI_studentized(base_g, results_g, studentized_results_g, alpha=alpha, alternative=alternative), 
                       data, g, lowess_linear_interp, frac=frac_invert, integration_precision=integration_precision)
    else:
        CI = compute_CI_studentized(base, results, studentized_results, alpha=alpha, alternative=alternative)
    return CI

def CI_percentile(data, statistic, R=int(1e5), alpha=0.05, smooth=False, alternative='two-sided', **kwargs):
    sample_stat = statistic(data)
    if hasattr(sample_stat, "__len__"):
        output_len = len(sample_stat)
    else:
        output_len = 1
    boot_sample = resample_nb(data, statistic, R=R, smooth=smooth, output_len=output_len, **kwargs)
    alpha_ptg = alpha*100
    if alternative == 'two-sided':
        CI = np.percentile(boot_sample, [alpha_ptg/2, 100 - alpha_ptg/2], axis=0).T
    elif alternative == 'less':
        CI = np.vstack((-np.inf * np.ones((output_len)), np.percentile(boot_sample, 100-alpha_ptg, axis=0))).T
    elif alternative == 'greater':
        CI = np.vstack((np.percentile(boot_sample, alpha_ptg, axis=0), np.inf * np.ones((output_len)))).T
    else:
        raise ValueError(f"alternative '{alternative}' not valid. Available: 'two-sided', 'less', 'greater'.")
    return CI

def CI_all(data, statistic, R=int(1e5), alpha=0.05, alternative='two-sided', coverage_iters=int(1e5), coverage_seed=42, avg_len=3, exclude=['studentized_vs', 'studentized_vs_smooth']):
    """
    Computes all CIs. 
    exclude: CIs to exclude. Available: percentile
                                        percentile_smooth
                                        bca
                                        bca_smooth 
                                        studentized
                                        studentized_smooth
                                        studentized_vs
                                        studentized_vs_smooth
    """
    specs = dict(percentile = (CI_percentile, {}),
                 percentile_smooth = (CI_percentile, dict(smooth=True)),
                 bca = (CI_bca, {}),
                 bca_smooth = (CI_bca, dict(smooth=True)),
                 studentized = (CI_studentized, {}),
                 studentized_smooth = (CI_studentized, dict(smooth=True)),
                 studentized_vs = (CI_studentized, dict(vs=True)),
                 studentized_vs_smooth = (CI_studentized, dict(vs=True, smooth=True))
                )
    specs = {k: v for k, v in specs.items() if k not in exclude} 
    
    CIs = defaultdict(list)
    for label, (func, kws) in tqdm(specs.items()):
        CI = func(data, statistic, R=R, alpha=alpha, alternative=alternative, **kws)
        CIs['CI'].append(label)
        if CI.shape[0] == 1 or CI.ndim == 1:
            CI = CI.squeeze()
            CIs['low'].append(CI[0])
            CIs['high'].append(CI[1])
        else:
            CIs['low'].append(CI[:, 0])
            CIs['high'].append(CI[:, 1])
    return conf_interval.CI_specs(pd.DataFrame(CIs).set_index('CI'), data, statistic, coverage_iters=coverage_iters, seed=coverage_seed, avg_len=avg_len)

def power_analysis_naive(data, statistic, low, high, N_values=np.array([5, 10, 25, 50, 100, 200]), R=int(1e4), seed=0):
    """
    Naive bootstrap power and sample-size calculation for accepting H0. 
    Computes violations on the low and high bound of H0.
    See Efron-Tshibirani: An introduction to the bootstrap,  p. 379-381.
    
    Returns: dataframe with index=N_values, columns=[low_fails, high_fails, power].
    """
    results = defaultdict(list)
    n = data.shape[0]
    if n not in N_values:
        N_values = np.sort(np.hstack((n, N_values)))
    for N in N_values:
        data_resampled = resample_nb(data, statistic, R=int(1e4), N=N, seed=seed)
        fail_low = (data_resampled < low).mean()
        fail_high = (data_resampled > high).mean()
        power = 1 - fail_low - fail_high
        results['violations-low'].append(fail_low)
        results['violations-high'].append(fail_high)
        results['power'].append(power)
    return pd.DataFrame(results, index=N_values)

def power_analysis(data, statistic, low, high, output_len=1, N_values=np.array([5, 10, 25, 50, 100, 200]), recenter=False, seed=0,
                   R=int(1e4), R_se=int(1e5), R_se_nested=int(1e3)):
    """
    Stable bootstrap power and sample-size calculation for accepting H0. 
    Computes violations on the studentized equivalent for the low and high bound of H0.
    Takes into account the variability in the original sample (size n):   bootstrap estimate, SE of the bootstrap estimate.
                                     and in the future sample (size N):   bootstrap estimate.
    See Efron-Tshibirani: An introduction to the bootstrap,  p. 381-384.
    
    Returns: dataframe with index=N_values, columns=[low_fails, high_fails, power].
    """
    n = data.shape[0]
    if n not in N_values:
        N_values = np.sort(np.hstack((n, N_values)))
    base = statistic(data)
    se_estimate = np.sqrt(np.diag(cov(resample_nb(data, statistic, seed=seed, R=R_se, output_len=output_len), 
                                      base,
                                      recenter=recenter)
                                 ))    
    low_studentized = (base - low) / se_estimate
    high_studentized = (base - high) / se_estimate
    
    # Estimation of SE (datasize = n). This is the computational bottleneck.
    data_n = resample_nb_X(data, R=R, seed=seed+1)
    se_estimate_n = np.empty((R, output_len))
    estimate_n = np.empty((R, output_len))
    for i, d_n in enumerate(tqdm(data_n)):
        estimate_n_i = statistic(d_n)
        nested_estimate_n = resample_nb(d_n, statistic, seed=seed+3+i, R=R_se_nested, output_len=output_len)
        estimate_n[i] = estimate_n_i
        se_estimate_n[i] = np.sqrt(np.diag(cov(nested_estimate_n, estimate_n_i, recenter=recenter))) 
        
    # Violations of studentized endpoints.
    results = defaultdict(list)
    for N in N_values:
        estimate_N = resample_nb(data, statistic, N=N, R=R, seed=seed+2, output_len=output_len)
        T = (estimate_n - estimate_N) / se_estimate_n           
        low_violations = (T >= low_studentized).mean()
        high_violations = (T <= high_studentized).mean()
        power = 1 - low_violations - high_violations
        results['violations-low'].append(low_violations)
        results['violations-high'].append(high_violations)
        results['power'].append(power)
        
    return pd.DataFrame(results, index=N_values)


"""
Hierarchical clustering. In the future will include more algorithms
"""
import pandas as pd
try:
    from scipy.cluster import hierarchy
    import matplotlib.pyplot as plt
except:
    pass
try:
    import plotly.express as px
    from .plots.plotly_utils import get_figure, set_multicategory_from_df
except:
    pass


def hierarchy_dendrogram(X, fontsize=30, out='data'):
    """
    Notes on the linkage matrix:
    This matrix represents a dendrogram, where elements
        1, 2: two clusters merged at each step,
        3: distance between these clusters,
        4: size of the new cluster - the number of original data points included.
    """
    is_df = isinstance(X, pd.core.frame.DataFrame)
    if is_df:
        labels = X.columns.to_list()
    else:
        labels = [*range(X.shape[1])] if len(X.shape) > 1 else None #int(-1 + np.sqrt(1 + 8*X.size)/2))] 
            
    fig = plt.figure(figsize=(8, 12))
    ax = plt.subplot(111)
    
    corr_linkage = hierarchy.ward(X.values if is_df else X)
    dendro = hierarchy.dendrogram(
        corr_linkage, labels=labels, ax=ax, leaf_rotation=90 #orientation="left"
    )
    if out == 'data':
        plt.close()
        return corr_linkage, dendro
    elif out == 'fig':
        if not is_df:
            plt.tick_params(axis='x', which='both', bottom=False, top=False, labelbottom=False)
        else:
            plt.xticks(fontsize=fontsize)
        plt.yticks(fontsize=fontsize)
        return fig
    else:
        raise ValueError(f"out '{out}' not valid. Available: 'data', 'fig'.")
    

def hierarhical_cluster_matrix(df, title, colorbar_x=0.9, ticksize=16, cmin=-1, cmax=1):
    _, dendro = hierarchy_dendrogram(df)
    order = dendro["leaves"]
    #corr = X.corr() if isinstance(X, pd.core.frame.DataFrame) else np.corrcoef(X)
    df_ordered = df.iloc[order,:].iloc[:, order]
    fig = px.imshow(df_ordered)
    fig.update_layout(margin=dict(l=0, b=30, r=60, t=10, pad=1), xaxis_tickfont_size=ticksize, yaxis_tickfont_size=ticksize,
                      coloraxis=dict(cmin=cmin, cmax=cmax, colorbar=dict(title_text=title, tickfont_size=16, title_font_size=20, x=colorbar_x)),
                      height=800, width=1200, font_size=20, hovermode=False)
    if isinstance(df.index, pd.core.indexes.multi.MultiIndex):
        set_multicategory_from_df(fig, df_ordered)
    return fig

def corr_cluster_matrix(df, corr='spearman', **kwargs):
    return hierarchical_cluster_matrix(df.corr(corr), corr.capitalize(), **kwargs)



from setuptools import setup, find_packages

setup(
    name='phdu',
    version='1.5.b7',
    author="Jorge Medina Hernández",
    author_email='medinahdezjorge@gmail.com',
    packages=find_packages("."),
    url='https://github.com/medinajorge/PhD-utils',
    download_url='https://github.com/medinajorge/PhD-utils/archive/refs/tags/v1.5-beta.tar.gz',
    description="Automatically store/load data in a tidy, efficient way. Includes functions for data visualization and analysis.",
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    keywords=['science', 'statistics', 'tidy', 'project organization', 'project', 'organization', 'path', 'storage'],
    classifiers = [
        "Programming Language :: Python :: 3",            
        "Development Status :: 4 - Beta",
        "License :: OSI Approved :: GNU Affero General Public License v3",
        "Operating System :: OS Independent",    
        "Topic :: Scientific/Engineering",
        "Topic :: Office/Business",
        "Intended Audience :: Science/Research",
    ],
    python_requires=">=3",
    install_requires=[
        'tidypath',     
        'numba',
        'numpy',
        'pandas',
        'scipy',
        'colour',
        'Pillow',
    ],
    extras_require={
        "all": ["matplotlib", "plotly", "kaleido", "statsmodels", "rpy2", "colorlover", "sklearn"],
        "basic": ["matplotlib", "plotly", "kaleido", "colorlover", "sklearn"],
        "matplotlib": "matplotlib",
        "plotly": ["plotly", "kaleido"],        
        "colorlover": "colorlover",
        'statsmodels': 'statsmodels',
        "r": "rpy2",
        "sklearn": "sklearn"
    },
)



"""
Helper functions.
"""

def sequence_or_stream(x, maxlen=None):
    """
    if maxlen is None => returns a stream
    else              => returns sequence of elements of x repeated up to the desired len.
    
    Attrs:
        - x (iterable)
        - maxlen (int|None)
    """
    x = list(x)
    x_len = len(x)
    if maxlen is None: # stream
        def stream():
            n = 0
            while True:
                yield x[n]
                n = (n + 1) % x_len
        return stream()
    else:
        mod = maxlen % x_len
        return x * (maxlen // x_len) + ([] if mod == 0 else x[:mod])


"""
Handle scripts.
- getopt_printer: prints getopt input in a readable way.
- getopt_reader:  automatically parse getopt opts using tidypath
- incompleted_programs_shell_script: Creates or prins sh script for programs that were not completed. To be used after 'delete_stdin_files'.
"""
import os
from tidypath import fmt

def getopt_printer(opts):
    """Prints getopt input in a readable way."""
    print('\n'.join(f'{opt} => {arg}' for opt, arg in (("Args", "Values"), *opts)))
    
def getopt_reader(opts, key_short, key_long, print_result=True):
    """
    Returns dict containing opts. args are parsed by tidypath.fmt.decoder.
    NOTE: If not sure about the parsing, do not use it. 
    """
    key_short_fmt = [f"-{k}" for k in key_short.split(":")[:-1]]
    key_long_fmt = [k[:-1] for k in key_long]
    env = {}
    for opt, arg in opts:
        for ks, kl in zip(key_short_fmt, key_long_fmt):
            if opt in (ks, f"--{kl}"):
                env[kl.replace("-", "_")] = fmt.decoder(arg)
                break
    if print_result:
        print(f"getopt_reader result:\n{env}")
    return env
    
def incompleted_programs_shell_script(script, save_path=None, save_sh=True, stdin_identifier=None, cores=1, memory=4, gpus=0, time="0:20",
                                      programs_dir="nuredduna_programmes/", sh_dir="nuredduna_programmes/", stdin_dir='stdin_files/', 
                                      ext='.out', lang="python", shell="bash"):
    """
    Creates or prins sh script for programs that were not completed. To be used after 'delete_stdin_files'.
    
    Attributes:
        - script:           program name
        - save_path:        None  =>  path = 'script_name'_incompleted.sh
                            path  =>  stores sh in path location
        - save_sh:          True  =>  saves sh in 'save_path'.
                            False =>  prints sh
        - stdin_identifier: if there are stdin files of multiple programs, select only the programs containing 'stdin_identifier'.
    """
    if not ext.startswith("."):
        ext = f".{ext}"
    if lang == 'python':
        script_name, script_ext = os.path.splitext(script)
        if script_ext != ".py":
            script = f"{script_name}.py"
        
    if gpus:
        hardware_specs = f"-c {cores} -m {memory} -g {gpus} -t {time}"
    else:
        hardware_specs = f"-c {cores} -m {memory} -t {time}"
        
    def run_command(args):
        stdin_name = os.path.join(stdin_dir, "_".join(args).replace("-", "").replace(" ", "-"))
        args_specs = " ".join(args)
        run_cmd = f'run {hardware_specs} -o {stdin_name}.out -e {stdin_name}.err "{lang} {script} {args_specs}"'
        return run_cmd
    
    stdin_dir_full = os.path.join(programs_dir, stdin_dir)
    programs = []
    for file in os.listdir(stdin_dir_full):
        if stdin_identifier is None or stdin_identifier in file:
            if os.path.splitext(file)[1] == ext:
                with open(os.path.join(stdin_dir_full, file)) as f:
                    args = []
                    for line in f.readlines()[1:]:
                        if "=>" in line:
                            arg = line.replace(" =>", "").replace("\n","")
                            args.append(arg)
                    if args:
                        programs.append(run_command(args))
                    
    shell_script = f"#!/bin/{shell}\n\n" + "\n".join(programs)

    if save_sh:
        if save_path is None:
            save_path = os.path.join(programs_dir, sh_dir if sh_dir != programs_dir else '', f"{script_name}_incompleted.sh")
        else:
            file, ext = os.path.splitext(save_path)
            if sh_dir not in file:
                file = os.path.join(sh_dir, file)
            if ext != ".sh":
                save_path = f"{file}.sh"
        with open(save_path, 'w') as f:
            f.write(shell_script)
            f.close()
    else:
        print(shell_script)
    return


"""
pandas utils
"""
import pandas as pd
import numpy as np
from functools import reduce     
from .stats.rtopy import resample
from .stats.test import permutation

def latex_table(df, index=False, **kwargs):
    """Pandas DataFrame -> Latex table."""
    col_format = "c" if isinstance(df, pd.core.series.Series) else "c"*len(df.columns)
    if index:
        col_format += "c"
    table_replacements = (("\\toprule", "\\toprule "*2),
                          ("\\bottomrule", "\\bottomrule "*2)
    )
    text_replacements = (("\\textbackslash ", "\\"),
                         ("\{", "{"), 
                         ("\}", "}"),
                         ("\$", "$"),
                         ("\_", "_"),
                         ("\\textasciicircum ", "^")
    )
    table_formatter = lambda x:  reduce(lambda a, kv: a.replace(*kv), table_replacements, x)
    text_formatter = lambda x: reduce(lambda a, kv: a.replace(*kv), text_replacements, x)
    formatter = lambda x: text_formatter(table_formatter(x))
    print(formatter(df.to_latex(index=index, column_format=col_format, **kwargs)))
    return

def expand_sequences(df, dt=1, maxlen=None):
    """
    Input: DataFrame. Each element is an array and all arrays start at the same time and have the same time step dt.
    
    Returns: MultiColumn DataFrame: (df.index,  (df.columns, time_steps))
    """
    if df.isna().values.any():
        if maxlen is None:
            maxlen = int(df.applymap(lambda x: x.size if isinstance(x, np.ndarray) else np.NaN).max().max())
        df_padded = df.applymap(lambda x: np.hstack((x, np.NaN*np.empty((maxlen-x.size)))) if isinstance(x, np.ndarray) else np.NaN*np.empty((maxlen)))
    else:
        if maxlen is None:
            maxlen = int(df.applymap(lambda x: x.size).values.max())
        df_padded = df.applymap(lambda x: np.hstack((x, np.NaN*np.empty((maxlen-x.size)))))
    df_padded_arr = np.stack([np.vstack(x) for x in df_padded.values]) # shape (df.shape[0], df.shape[1], time_steps)
    return pd.DataFrame(df_padded_arr.reshape((df.shape[0], -1)), 
                        index = df.index, 
                        columns = pd.MultiIndex.from_product([df.columns, dt*np.arange(maxlen)]))

def tuple_wise(*dfs):
    """Returns dataframe where each element is a tuple containing the elements from other dataframes."""
    df = dfs[0]
    assert all(df.index.intersection(df2.index).size == df.shape[0] for df2 in dfs[1:])
    assert all(df.columns.intersection(df2.columns).size == df.shape[1] for df2 in dfs[1:])
    return pd.DataFrame(np.rec.fromarrays(tuple(df.values for df in dfs)).tolist(), 
                        columns=df.columns,
                        index=df.index)

def vstack_wise(*dfs):
    """
    Attributes: Dataframes where elements are arrays with equal length and have same indices and columns.
    
    Returns: DataFrame where df_ij = np.vstack((df1_ij, df2_ij, ...))
    """
    R = np.rec.fromarrays(tuple(df.values for df in dfs))
    df1_df2 = pd.Series([np.vstack(i) if all(isinstance(j, np.ndarray) for j in i) else np.NaN for i in R.flatten()], dtype=object,
                        index=dfs[0].stack(dropna=False).index).unstack()
    return df1_df2

def column_diffs(df, mode="to", mod_col=lambda x: "".join(np.array([*x])[[0, -1]])):
    """
    mode: '-'   (col2 - col1)
          'to' r'$col1 \to col2
    """
    N = df.shape[1]
    data = {}
    if mode == "-":
        label = lambda col1, col2: f"{col2} - {col1}"
    elif mode == "to":
        if callable(mod_col):
            label = lambda col1, col2: r"$\Huge {{{}}} \to {{{}}}$".format(mod_col(col1.replace(" ", "\ ")), mod_col(col2.replace(" ", "\ ")))            
        else:
            label = lambda col1, col2: r"$\Huge {{{}}} \to {{{}}}$".format(col1.replace(" ", "\ "), col2.replace(" ", "\ "))
    for i, col1 in enumerate(df.columns):
        if i < N:
            for col2 in df.columns[i+1:]:
                data[label(col1, col2)] = df[col2] - df[col1]
    return pd.DataFrame(data)

def CI_by_col(df, stat, return_sample_stat=True, fillna=None, **kwargs):
    CI = pd.Series({col: resample.CI_bootstrap(df[col].dropna().values, stat=stat, **kwargs) for col in df.columns})
    if fillna is not None:
        def aux(x):
            x[np.isnan(x)] = fillna
            return x
        CI = CI.apply(aux)
    if return_sample_stat:
        return CI, getattr(df, stat)()
    else:
        return CI
    
def permtest_by_col(df, alternative, stat="mean", paired=True, diff=True, label="c", **kwargs):
    """
    Test directionality accross columns. 
        col_i  (direction in [<, >, !=])  col_j     
        
    Attributes:
        alternative:    'less', 'greater', 'two-sided'.
    """
    P = {}
    N = df.shape[1]
    is_paired = "paired" if paired else "not_paired"
    is_diff = "diff" if diff else ""
    for i in range(N):
        for j in range(i+1, N):
            yi, yj = df.iloc[:, [i,j]].dropna().values.T
            P[r"$\Huge {}_{{{}{}}}$".format(label, i+1, j+1)] = getattr(permutation, f"permutation_test_2sample_{is_paired}_{is_diff}{stat}")(yi, yj, alternative=alternative, **kwargs)
    return pd.Series(P)


import numpy as np
from numba import njit

def simpson3oct_vec(f, a, b, pre, *args, relative=True):
    """
    Simpson 3 octaves method.
    Attrs:
        - f:          integrand
        - a:          lower end of integration
        - b:          upper end of integration
        - pre:        precision
        - relative:   precision w.r.t. the order of magnitude of the integration
                      (False: absolute precision)
    """
    num = 3
    In = -1
    In1 = 100.
    if relative:
        condition = lambda In, In1: abs((In-In1) / In1) > pre
    else:
        condition = lambda In, In1: abs(In - In1) > pre
    while condition(In, In1):
        xgrid = np.linspace(a, b, num+1)
        ygrid = f(xgrid, *args)
        if np.isnan(ygrid).all(): #restart
            In = 0. 
            In1 = 100.
        else:            
            In1 = In
            h_3oct = 3*(b-a) / (8*num) # num + 1 edges, num intervals 
            h_6oct = 2 * h_3oct
            h_9oct = 3 * h_3oct
            H = np.hstack((h_3oct, np.repeat(np.array([h_9oct, h_9oct, h_6oct])[None], 1 + num//3, axis=0).ravel()[:(num-1)], h_3oct))
            In = (H * ygrid).sum()
        num += 3
    return In, num-3

@njit
def simpson3oct(f, a, b, pre, *args):
    """
    Simpson 3 octaves method.
    Attrs:
        - f:   integrand
        - a:   lower end of integration
        - b:   upper end of integration
        - pre: precision
    """
    num = 3
    In = 0.
    In1 = 100.
    while abs(In - In1) > pre:
        In1 = In
        h = (b-a) / num
        In = 0.
        n = 1
        for k in np.linspace(a, b, num+1):
            if k == a:
                In = In + (3/8.)*h*f(k, *args)
            if k == b:
                In = In + (3/8.)*h*f(k, *args)
            elif n == 2:
                In = In + (9*h/8.)*f(k, *args)
            elif n == 3:
                In = In + (9*h/8.)*f(k, *args)
            elif n == 4:
                In = In + (6*h/8.)*f(k, *args)
            n += 1
            if n == 5:
                n = 2
        num += 3
    return In, num-3

@njit
def trapecios(f,a,b,pre, *args):
    num = 1
    In = 0.
    In1 = 100.
    while abs(In - In1) > pre:
        In1 = In
        h = (b-a) / num
        In = 0.
        for k in np.linspace(a, b, num+1):
            if k == a:
                In = In + (h/2.)*f(k, *args)
            elif k == b:
                In = In + (h/2)*f(k, *args)
            else:
                In = In + h*f(k, *args)
        num += 1
    return In, num-1


from tidypath import savedata, savefig, add_arg, modify_arg, delete_arg
from . import script_fmt
from . import np_utils
from . import pd_utils
from . import storage
from . import plots
from . import stats
from . import decomposition
from . import clustering
from .stats import bootstrap, conf_interval
from .stats.rtopy import resample
from .stats.test import permutation
from .storage import delete_stdin_files
from .script_fmt import getopt_printer, incompleted_programs_shell_script
